from actappliance.act_errors import act_errors, ACTError
from actappliance.permissions import AppliancePermissions
from actappliance.connections import ApplianceSsh, ApplianceRest
from bitpermissions import PermissionException
import logging
from paramiko import AuthenticationException
import warnings


appl_types = ['cds', 'sky', 'psky', 'skyng']
appl_types_str = ', '.join(appl_types[:-1]) + ', or ' + appl_types[-1]

conn_types = {'ssh': 'cli',
              'cli': 'cli',
              'rest': 'rest'}
conn_types_keys = list(conn_types.keys())
conn_types_str = ', '.join(conn_types_keys[:-1]) + ', or ' + conn_types_keys[-1]
del conn_types_keys


class Appliance:
    """
    An appliance represents an Actifio uds product like cds or sky.

    As a general rule use cmd method for all udsinfo and udstask commands. This allows them to be changed dynamically
    at runtime. If you are sending a hidden rest command use rest and if you are sending a command that can't be sent
    over rest use raw.

    Usage:
     a = Appliance(ip_address='172.27.4.0')
     r = a.cmd('udsinfo lshost -filtervalue hostname=dog&hosttype=puppey')
     hostid = r.parse(k='id')

    The API for this object is aided by dynamic_keywords. If we call a method with a name matching the conversion
    table in keyword_prefixes, we will attempt to build a simple call and check method. This avoids writing a lot of
    boilerplate and filters all calls through a specific naming convention and methods, so it can be monkeypatched if
    a feature changes or a bug is found.
    """

    def __init__(self, hostname=None, ip_address=None, connection_type='rest', appliance_type=None, version=None,
                 ssh_args=None, rest_params=None, *, logger=None):
        """
        Setup config info as well as permissions

        ssh_args should normally have username, password, port as keys.
        rest_params should normally have name, password, vendorkey as keys.

        :param hostname: DNS name of machine to connect to (hostname or ip_address required)
        :param ip_address: Ip address of machine to connect to (hostname or ip_address required)
        :param connection_type: Default way to connect to appliance (options: cli, rest)
        :param appliance_type: Category of appliance
        :param ssh_args: Dictionary of ssh connection args to send to paramiko's connect method
        :param rest_params: Dictionary of connection params to send alongside RESTful requests
        """
        # These permissions are to stop users from hitting unexpected failures. Not for security.
        self.perms = AppliancePermissions()

        self._connection_type = self._version = self._appliance_type = None

        self.logger = logger or logging.getLogger(__name__)

        # Get system info
        self.hostname = hostname
        self.ip_address = ip_address
        self.system = ip_address or hostname
        if not self.system:
            self.logger.info('Ip address or hostname required to send commands')

        # Get connection info
        self.ssh_args = ssh_args if ssh_args is not None else {}
        self.rest_params = rest_params if rest_params is not None else {}

        # Get optional appliance type
        self.appliance_type = appliance_type
        if appliance_type:
            self.update_class(appliance_type)
        self.version = version

        # Ensure connection input is valid
        self.connection_type = connection_type

        # Make connections for commands
        self._make_connections()

    @staticmethod
    def _get_appl_class(appl_type):
        return appl_classes.get(appl_type.lower(), Appliance)

    def update_class(self, appl_type=None):
        appl_type = self.appliance_type if appl_type is None else appl_type
        self.__class__ = self._get_appl_class(appl_type)

    @property
    def connection_type(self):
        return self._connection_type

    @connection_type.setter
    def connection_type(self, value):
        connection_type = value.lower()
        if connection_type in conn_types:
            self._connection_type = conn_types[connection_type]
        else:
            raise ValueError("'connection_type' must be {}".format(conn_types_str))

    def _make_connections(self):
        self.rest_conn = ApplianceRest(self.system, self.rest_params)
        self.ssh_conn = ApplianceSsh(self.system, self.ssh_args)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.teardown()

    def teardown(self):
        self.rest_conn.disconnect()
        self.ssh_conn.disconnect()

    @property
    def appliance_type(self):
        if not self._appliance_type:
            self._appliance_type = self.get_appliance_type()
        return self._appliance_type

    @appliance_type.setter
    def appliance_type(self, value):
        self._appliance_type = value.lower() if value is not None and value.lower() in appl_types else None

    def get_appliance_type(self):
        """Find the appliance's type. ex. sky, cds, skyng"""
        appliance_type = None
        try:
            self.logger.debug("Searching for appliance type over ssh.")
            out, err, rc = self.raw('rpm -qa | grep actsys | grep -ho -e svc -e skyng -e psky -e sky', statuses=[0])
            try:
                actsys_type = out[0].lower()
            except IndexError:
                raise ValueError("rpm actsys didn't return a meaningful type.")
            appliance_type = actsys_type.replace('svc', 'cds')
            if appliance_type not in appl_types:
                msg = "Appliance type {} from rpm -qa isn't {}".format(appliance_type, appl_types_str)
                self.logger.exception(msg)
                raise ValueError(msg)
        except (AuthenticationException, ValueError, PermissionException):
            self.logger.debug("Failed to get appliance type over ssh, searching with RESTful API.")
            r = self.rest('udsinfo lsversion')
            r.raise_for_error()
            for di in r['result']:
                for v in di.values():
                    value = v.lower()
                    if value in appl_types:
                        appliance_type = value
                        break

        appliance_type = appliance_type.replace('psky', 'skyng')
        if appliance_type not in appl_types:
            raise RuntimeError("Invalid appliance type {0} found".format(appliance_type))
        self.logger.info("Appliance type is {}.".format(appliance_type))

        return str(appliance_type)

    @property
    def current_version(self):
        return self.get_version()

    @property
    def version(self):
        """
        Version is not re-evaluated once found and we don't check if you set the value incorrectly.

        Use `current_version` if you need lazy evaluation.
        """
        if not self._version:
            self._version = self.get_version()
        return self._version

    @version.setter
    def version(self, value):
        self._version = value

    def get_version(self):
        r = self.cmd('udsinfo lsversion')
        version = r.parse(k='version', m_k='component', m_v='srv-revision')
        if not version:
            self.logger.info('This system looks older and may call srv-revision psrv-revision, attempting to grab.')
            version = r.parse(k='version', m_k='component', m_v='psrv-revision')
        if not version:
            raise RuntimeError("No version found.")
        self.logger.info('srv-version is {}'.format(version))
        return version

    # Helper methods for easy execution

    def cmd(self, operation, **update_cmds):
        """
        Connect to the machine using the classes connection type and send the given command.

        If an argument needs to be sent in the operation and not in kwargs with argument make sure it is first in the
        series of commands.

        Note: There is an abstraction leak in how the API interprets strings as booleans, you should pass in true bools
              so that it works with CLI and REST.
              For example appl_obj.cmd('udstask mountimage', nowait='True') would not wait on rest connection type,
              but would be interpreted as a string by cli and cause an error. So, for consistency always pass in bools
              as bools (True/False)

        :param operation: CLI-like command to send
        :param update_cmds: add or overwrite commands in the operation. Will not add if value is none.
        :return: ActResponse object
        """
        self.perms.has_cmd()
        if self.connection_type == "rest":
            return self.rest(operation, **update_cmds)
        elif self.connection_type == "cli":
            return self.cli(operation, **update_cmds)
        else:
            raise RuntimeError("Couldn't find suitable cmd method based on connection type. {} "
                               "provided".format(self.connection_type))

    def rest(self, operation, **update_cmds):
        """
        Use when you explicitly want to connect over RESTful API.

        :param operation: CLI-like command to be sent. ex. udsinfo lshost
        :param update_cmds: keyword arguments to be sent with command
        :return:
        """
        self.perms.has_rest()

        return self.rest_conn.cmd(operation, **update_cmds)

    def cli(self, operation, **update_cmds):
        """
        Use when you explicitly want to connect over ssh.

        :param operation: CLI-like command to be sent. ex. udsinfo lshost
        :param update_cmds: keyword arguments to be sent with command
        :return: ActResponse object
        """
        self.perms.has_ssh()

        return self.ssh_conn.cmd(operation, **update_cmds)

    def raw(self, command, *args, **kwargs):
        """
        Use when you explicitly want to connect over ssh and send raw ssh commands.

        This houses the functionality unlike rest method because I want to avoid passing around the client to retain
        channels.
        :param command:
        :return: stdout, stderr, exit status
        """
        self.perms.has_ssh()

        return self.ssh_conn.raw(command, *args, **kwargs)

    def ui_call(self, method, url_suffix, params=None, data=None, **kwargs):
        """
        Used to send Actifio desktop rest calls.

        :param method: type of rest call to make. ie. GET, POST
        :param url_suffix: What follows actifio in the UI call. i.e. https://<some_hostname>/actifio/<url_suffix here>
        :return:
        """
        self.perms.has_rest()
        return self.rest_conn.ui_call(method, url_suffix, params, data, **kwargs)

    def new_user(self, new_name, new_password=None):
        """
        This is an idempotent way to select and sign in as a new user with default rights.

        :param new_name: New username to create or find
        :param new_password: New password to set or use
        :return: New Appliance object with the new user information based off of the current instance
        """
        warnings.warn("This shouldn't be provided by this module, replicate externally.", DeprecationWarning)
        if new_password is None:
            new_password = self.rest_params.get('password')

        try:
            r = self.cmd('udstask mkuser', name=new_name, password=new_password)
            r.raise_for_error()
        except act_errors[10006]:
            # This assumes the password of the created user is new_password, so we have to check the password works
            self.logger.info('User {} already existed.'.format(new_name))
            check_login = True
        else:
            # we don't need to check_login if exceptions aren't hit
            check_login = False

        # Remove the ssh permissions
        new_perms = self.perms
        new_perms.ssh = False
        # set the new params
        new_rest_params = self.rest_params.copy()
        new_rest_params['name'] = new_name
        new_rest_params['password'] = new_password

        new_args = []
        for arg in [self.hostname, self.ip_address, self.connection_type, self.appliance_type, self.version,
                    self.ssh_args]:
            try:
                new_args.append(arg.copy())
            except AttributeError:
                new_args.append(arg)

        new_appliance = Appliance(*new_args, rest_params=new_rest_params)
        new_appliance.perms = new_perms

        # If we found an existing user, make sure our password works
        if check_login:
            r = new_appliance.rest('udsinfo lsversion')
            try:
                r.raise_for_error()
            except ACTError:
                raise RuntimeError("User {} already existed and the provided password was incorrect.".format(new_name))

        return new_appliance


class Sky(Appliance):
    pass


class CDS(Appliance):
    pass


appl_classes = {
    'cds': CDS,
    'sky': Sky
}



from abc import ABC, abstractmethod
import atexit
from concurrent.futures import ThreadPoolExecutor as Executor
from concurrent.futures import TimeoutError
import logging
import shlex
import warnings

from actappliance.models import ActCmd


class ApplianceConnection(ABC):

    def __init__(self, system, connect_kwargs, call_timeout=6 * 60 * 60):
        self.logger = logging.getLogger(__name__)

        self.system = system
        self.timeout = call_timeout
        self.connect_kwargs = connect_kwargs if connect_kwargs is not None else {}

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.disconnect()
        atexit.unregister(self.disconnect)

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    @abstractmethod
    def _auto_connect(self):
        pass

    @abstractmethod
    def call(self, act_cmd):
        """Call that gets the unfiltered connections response"""
        self._auto_connect()

    def _timeout(self, fnc, *args, **kwargs):
        with Executor(max_workers=1) as p:
            f = p.submit(fnc, *args, **kwargs)
            try:
                return f.result(timeout=self.timeout)
            except TimeoutError:
                timeout = "Call timed out after {} seconds.".format(self.timeout)
                self.logger.exception("function: {}\n" +
                                      "args: {}\n" +
                                      "kwargs: {}".format(self.timeout, fnc, args, kwargs))
                raise TimeoutError(timeout)

    def timeout_call(self, act_cmd):
        return self._timeout(self.call, act_cmd)

    def cmd(self, operation, **update_cmds):
        """Command that makes a call with standardized input and output so it's the same across connections"""
        ac = ActCmd(conn_object=self)
        ac.cmd_input(operation, **update_cmds)
        self.timeout_call(ac)
        return ac.response

    # TODO: Move this to ActInputs or make a function somewhere else
    def append_filtervalue(self, command, **update_cmds):
        """
        Takes a operation and an update_cmds dict and adds the new filtervalues to the existing one instead of
        overwriting.

        :param command:
        :param update_cmds:
        :return: operation and correctly appended update_cmds
        """
        warnings.warn("This is being removed from connection classes.", DeprecationWarning)
        # handle no input
        if 'filtervalue' not in update_cmds:
            # return empty string so that update will function without TypeError
            return ''
        filtervalue_append = update_cmds.pop('filtervalue')
        if filtervalue_append:
            s_command = shlex.split(command)
            for word in enumerate(s_command):
                if word[1] == '-filtervalue':
                    position = word[0]
                    update_cmds['filtervalue'] = "{}&{}".format(s_command[position + 1], filtervalue_append)
            try:
                update_cmds['filtervalue']
            except KeyError:
                self.logger.info('No existing filtervalue found returning update_cmd filtervalue only')
                update_cmds['filtervalue'] = filtervalue_append
        return update_cmds

    class ActCmd(object):
        """Class to house attributes for a single cmd"""

        def __init__(self, operation=None, **update_cmds):
            self.operation = operation
            self.update_cmds = update_cmds
            self.call_result = None  # This holds result before it's processed in post_cmd if required
            self.result = None  # This is the temporary holder for the REST-ful response

        def __repr__(self):
            """This will be None if post_cmd gets called, check the ActResponse dictionary for the values"""
            return str(self.result)



import requests
import json
import atexit

from actappliance.models import ActResponse
from actappliance.connections.connection import ApplianceConnection

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util import Retry
# suppress warnings from verify=False
from requests.packages.urllib3.exceptions import SecurityWarning
requests.packages.urllib3.disable_warnings(SecurityWarning)
# This is required if we want to support old versions of py2
# from requests.packages.urllib3.exceptions import SNIMissingWarning
# requests.packages.urllib3.disable_warnings(SNIMissingWarning)


requests_timeout = (121, 241)


class ApplianceRest(ApplianceConnection):

    def __init__(self, *args, **kwargs):
        """
        :param system: Appliance to connect to (ip or dns name)
        :param connect_kwargs: Dictionary of connect_kwargs to send to connect; ex. {'name': 'ex', 'password': 'ex', 'vendorkey': 'ex'}
        :param call_timeout: Maximum time to spend attempting an individual call
        """
        super(ApplianceRest, self).__init__(*args, **kwargs)

        self.connect_kwargs.setdefault('vendorkey', 'novend')

        # Set in method 'connect'
        self.s = self.sid = None

    def connect(self):
        """Connect and get a sessionid from the appliance."""
        # Ensure we have everything we need to connect before attempting
        for input_ in ['name', 'password', 'vendorkey']:
            if input_ not in self.connect_kwargs:
                raise ValueError("Missing parameter {}, cannot login without it.".format(input_))

        self.s = requests.Session()
        # 20 total seconds of waiting: {backoff factor} * (2 ^ ({number of retries} -1))
        adapter = HTTPAdapter(max_retries=Retry(connect=5, backoff_factor=1.25))
        self.s.mount("https://" + self.system, adapter)

        self.logger.info("Logging in as user '{}'.".format(self.connect_kwargs.get('name', '<no name found>')))

        result = self.s.post("https://" + self.system + "/actifio/api/login", verify=False, params=self.connect_kwargs,
                             timeout=requests_timeout)
        result.raise_for_status()
        result_j = result.json()
        if 'errorcode' in result_j:
            raise ValueError("Failed to connect: {}".format(result_j))

        self.sid = result_j['sessionid']
        self.logger.info('Sessionid {} acquired.'.format(self.sid))

        self.s.keep_alive = False

        # Sign out after execution
        atexit.register(self.disconnect, sid=self.sid)

    def disconnect(self, sid=None):
        """Logout of an actifio rest session."""
        if sid is None:
            sid = self.sid

        # Only try to disconnect if we have a sid. It takes a long time to retry connection to a non-existent machine.
        if sid:
            self.logger.debug("Logging out of session {}".format(sid))
            url = ("https://{}/actifio/api/logout?sessionid={}".format(self.system, sid))
            try:
                self.s.post(url, verify=False, timeout=requests_timeout)
            except requests.ConnectionError:
                # Don't raise connection error. It's safe to let it timeout.
                self.logger.info("Unable to logout of session, it may have already expired.")
            self.s.close()

        # set sid to None for the case where it was set as self.sid
        self.sid = None if self.sid is sid else self.sid

    def _auto_connect(self):
        if not self.sid:
            self.connect()

    def call(self, act_cmd):
        super(ApplianceRest, self).call(act_cmd)

        r = self._protect_sid(**act_cmd.input.rest_kwargs)
        try:
            r.raise_for_status()
        except requests.HTTPError as e:  # NOQA: keep this object for debugging
            self.logger.debug("Bad HTTP Status {}.\nResponse: {}\nUrl {}".format(r.status_code, r.json(), r.url))
            # Let the user handle errors

        act_cmd.output.result = r.json()

    def _protect_sid(self, method, url, params, verify=False, **kwargs):
        """Resend the operation if it looks like the sessionid has been dropped (default times out at 30 minutes)."""
        try:
            r = self.s.request(method, url, params=params, verify=verify, **kwargs)
            self.logger.info('Request url: {}'.format(r.url))
            self.logger.debug(r.content)
            r.raise_for_status()
        except requests.HTTPError as e:
            # errorcode 10020 is User is not logged in
            # {'errorcode': 10020, 'errormessage': 'User is not logged in'}
            try:
                errorcode = r.json()['errorcode']
            except KeyError:
                self.logger.debug(r.json())
                self.logger.exception("Status code was bad, but no errorcode could be found!")
                raise
            if e.response.status_code == 500 and errorcode == 10020:
                self.logger.exception('Sessionid was invalid, re-authenticating and sending again.')
                self.logger.debug(r.json())
                self.logger.debug("Failed sessionid is '{}'".format(self.sid))
                self.connect()
                params['sessionid'] = self.sid
                self.logger.debug("New sessionid is '{}'".format(self.sid))
                r = self.s.request(method, url, params=params, verify=verify, **kwargs)

            # Don't re-raise let the user handle all other errors

        return r

    def ui_call(self, method, url_suffix, params=None, data=None, **kwargs):
        """
        Used to send Actifio desktop rest calls.

        :param method: type of rest call to make. ie. GET, POST
        :param input_dictionary: python object to be converted to json
        :param url_suffix: What follows actifio in the UI call. i.e. https://<some_hostname>/actifio/<url_suffix here>
        :return: ActResponse object of request
        """
        if not self.sid:
            self.connect()

        if params is None:
            params = {}
        url = "https://{}/actifio/{}".format(self.system, url_suffix)
        params['sessionid'] = self.sid
        r = self._protect_sid(method, url, params=params, data=data, timeout=requests_timeout, **kwargs)
        r.raise_for_status()
        # strip the unparseable parts of the response that the UI framework needs and convert to dictionary
        answer = json.loads(r.text[1:-1])
        self.logger.debug(answer)
        if answer.get('status') != 'ok':
            ar = ActResponse(answer)
            ar.raise_for_error()
        return answer



import numbers
import logging
import shlex
import re
from typing import List
from actappliance.act_errors import act_errors, ACTError

"""
This module contains the primary objects that power actappliance.
"""

logger = logging.getLogger('models')
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)


class ActResponse(dict):
    """Dictionary class that helps parsing and raising errors on public Actifio API responses."""

    def __init__(self, *args, actifio_command=None, **kwargs):
        """

        :param args: dictionary args
        :param actifio_command: Deprecated
        :param kwargs: dictionary kwargs
        """
        self.command = actifio_command
        super(ActResponse, self).__init__(*args, **kwargs)
        self.errorcode = None
        self.errormessage = None

    def raise_for_error(self):
        """Raise ACTError if one occurred"""
        try:
            self.errorcode = self['errorcode']
        except KeyError:
            logger.debug('No errorcode found in ActResponse.')

        try:
            self.errormessage = self['errormessage']
        except KeyError:
            # Handle UI responses
            try:
                self.errormessage = self['msg']
            except KeyError:
                logger.debug('No errormessage found in ActResponse.')

        if self.errorcode or self.errormessage:
            # Log the input data if added to the object
            try:
                logging.info(self.input)
                logging.debug(repr(self.input))
                logging.debug(f"SSH kwargs for failed cmd:\n{self.input.ssh_kwargs}")
                logging.debug(f"REST kwargs for failed cmd:\n{self.input.rest_kwargs}")
            except AttributeError:
                # No inputs loaded to add for debugging
                pass

            if self.errorcode in act_errors:
                raise act_errors[self.errorcode](self)
            else:
                raise ACTError(self)

    def parse(self, k=None, m_k=None, m_v=None, index=0, warnings=True):
        """
        Takes appliance api response and returns a desired value or dictionary.

        It will first search outside the result for the key. Then it will search inside the result.
        If only a key is given it will return the value of that key from the desired index (default 0).
        If given a match_key and match_value it will search for the first matching dictionary inside the result list
        and return the key from that dictionary.

        Index is ignored without warning if match_key and match_value are provided.

        It is outside the scope of this method to return multiple dictionaries. Parsing multiple outputs shouldn't use
        this method.

        :param self: python object response from appliance api (most commonly json decoded from method cmd)
        :param k: the key; corresponding value will be returned
        :param m_k: match key to find the indexes
        :param m_v: match value to find the indexes
        :param index: index of the list of dictionaries to return
        :param warnings: boolean of whether to raise warning or not
        :return: string if key provided; dictionary if key is not provided
        """
        # Is the result empty?
        try:
            if not self['result']:
                value = self._get_external_value(k)
                if value:
                    logger.info("External value {} found, but will not return from parse.".format(value))
                if warnings:
                    logger.warning('No result found. Returning nothing from parse.')
                return

        # safety net against strange responses
        except (IndexError, TypeError):
            logger.exception('Are you sure you wanted to parse this? You gave no key and there is no result.')
            raise
        except KeyError:
            logger.info('No result dictionary found.')
            return

        # Is the result just an id or response?
        if isinstance(self['result'], (str, numbers.Integral)):
            logger.debug('result looks like an id or name; returning immediately')
            return self['result']

        # Is the result just info on a single object?
        if isinstance(self['result'], dict):
            if k is None:
                return self['result']
            else:
                return self['result'][k]

        # Is the result several objects?
        if isinstance(self['result'], list):
            # get matching indexes
            if m_k is not None and m_v is not None:
                indexes = []
                for i, dictionary in enumerate(self['result']):
                    if dictionary.get(m_k) == m_v:
                        indexes.append(i)
                if not indexes:
                    logger.info('Nothing found matching {} and {}'.format(m_k, m_v))
                    return
                index = indexes[0]
            if k is None:
                # return dictionary
                return self['result'][index]
            else:
                # Return value
                try:
                    return self.get('result')[index][k]
                except IndexError:
                    logger.error("Index {} doesn't exist. If you manually entered an index, try using matching key "
                                 "and value instead.".format(index))
                    raise
                except KeyError:
                    return self._get_external_value(k)

    def _get_external_value(self, k):
        # Is the response outside of result?
        if k is not None:
            if k in self:
                value = self[k]
                if k in ['status', 'result']:
                    logger.info('There is a top level matching key {}. Using this variable is not a supported use of '
                                'parse.'.format(k))
                    logger.debug('Top level key {} is: {}.'.format(k, value))
                    return
                logger.debug("The top level value for {} is: {}".format(k, value))
                return value
        return


class Act_put:
    """Base class for Act*put i.e. ActInput, ActOutput"""
    _ssh_delim = '|MiLeD|'


class ActInput(Act_put):
    """
    Ephemeral ActInputs handler

    It is important that if there is an 'argument' in the operation that it is the first input.

    Without having outside information on every command it would be impossible for us to know when a switch has no input
    and when it is followed by an argument.

    Example:
      udsfakecommand -some_arg some_value
    In the above is some_arg a 'switch' which means that it is flagging on and some_value is the argument? Or is
    some_arg a key which some_value is the input for? We cannot figure out what parameters are switches and which expect
    a value. However, we can still parse all possible inputs as long as the argument is first.
    """
    parse_op_props = ['http_method', 'base_path', 'action', 'operation_params', 'action']

    def __init__(self, operation, *, conn_object=None, update_cmds: dict=None):
        """

        :param operation: The full operation to be converted, with any argument as the first input
                             example: "udsinfo lshost 1234 -filtervalue some=thing&also=this"
        :param conn_object: The connection object which you will make the call with
        :param update_cmds: Input or update key/value information sent with the command, this always has priority
        """
        self.operation = operation
        self.update_cmds = {} if update_cmds is None else update_cmds
        self.conn_object = conn_object

        for pub_prop in ActInput.parse_op_props:
            setattr(self, '_' + pub_prop, None)
        self._act_method = None

    def __repr__(self):
        return f"{self.__class__.__name__}({self.operation}, update_cmds={self.update_cmds})"

    def __str__(self):
        return f"system: '{self.system}'\narguments: {self.updated_cmd_inputs}"

    def parse_operation(self):
        op_part = self.operation.partition(' ')
        self.act_method = op_part[0]
        body = self.operation.partition(' ')[2]
        shlexd_body = shlex.split(body)
        self.action = self._get_action(shlexd_body)
        self.operation_params = self._get_operation_params(shlexd_body)

    @property
    def updated_cmd_inputs(self):
        updated = self.operation_params
        try:
            updated['sessionid'] = self.conn_object.sid
        except AttributeError:
            logger.debug("Connection object doesn't have a sid.")
        return {**self.operation_params, **self.update_cmds}

    @property
    def ssh_command(self):
        """This is the arg for a paramiko exec_command call"""
        ssh_command = self.format_ssh_command()
        logger.debug(f"SSH command: '{ssh_command}'")
        return ssh_command

    @property
    def ssh_kwargs(self):
        """These are the inputs for a paramiko exec_command call"""
        return {'command': self.ssh_command}

    @property
    def rest_kwargs(self):
        """These are the inputs for a Session.request call"""
        request_kwargs = {
            'method': self.http_method,
            'url': self.url,
            'params': self.updated_cmd_inputs
        }
        logger.debug(f"Request inputs: '{request_kwargs}'")
        return request_kwargs

    @property
    def systemless_rest_kwargs(self):
        """Used for when you want to see rest_kwargs without a specific target system

        This won't actually work if you attempt to send it because the url won't resolve.
        """
        request_kwargs = {
            'method': self.http_method,
            'url': "/actifio/api/" + self.base_path + "/" + self.action,
            'params': self.updated_cmd_inputs
        }
        return request_kwargs

    @property
    def system(self):
        try:
            return self.conn_object.system
        except AttributeError:
            logger.exception("Need a connection object with system to generate input!")
            raise

    @property
    def url(self):
        return "https://" + self.system + "/actifio/api/" + self.base_path + "/" + self.action

    @property
    def cli_method(self):
        return '/act/bin/' + self.act_method

    @staticmethod
    def _kwargs_to_cli(kwargs):
        cli_args = ''
        # add parameters to body
        for k, v in kwargs.items():
            # don't pass None connect_kwargs
            if v is not None:
                # handle special case argument
                if k.lower() == 'argument':
                    cli_args += ' {}'.format(v)
                    continue
                parameter = "-{}".format(k)
                # handle parameterless switches
                if v is True:
                    value = ''
                else:
                    value = shlex.quote(v)
                option = " ".join([parameter, value])
                cli_args += f' {option}'
        return cli_args

    def format_ssh_command(self):
        """
        Takes a functional cli operation and appends rest like inputs.

        ex.
          self.cmd('udsinfo lshost', filtervalue='ostype=Linux')
        returns:
          "/act/bin/udsinfo lshost -filtervalue 'ostype=Linux'"
        """
        cli_args = self._kwargs_to_cli(self.updated_cmd_inputs)

        if self.act_method == 'udsinfo' and self.action.startswith(('ls', 'list')):
            cli_args += " -delim '{}'".format(self._ssh_delim)
        full_operation = " ".join((self.cli_method, self.action, cli_args))
        return full_operation

    @staticmethod
    def _parm_list_to_dict(params_list: List):
        # create param dictionary
        params_dict = {}
        k = None
        v = None

        def multi_arg_catch(arg_value, args=[]):
            if not args:
                args.append(arg_value)
                return arg_value
            logger.debug(f"Multiple argument values '{args + [arg_value]}' found in operation.")
            raise ValueError("Multiple arguments in operation.")

        for i in reversed(params_list):
            if i.startswith('-'):
                k = i[1:]
            else:
                if v is None:
                    v = i
                else:
                    # handle sequential values
                    params_dict['argument'] = multi_arg_catch(v)
                    v = i
            if k:
                if v:
                    params_dict[k] = v
                    k = None
                    v = None
                else:
                    params_dict[k] = 'true'
                    k = None
        # grab the argument if it's the first parameter
        if v is not None:
            params_dict['argument'] = multi_arg_catch(v)
        return params_dict

    @staticmethod
    def _get_action(shlexd_body):
        try:
            action = shlexd_body[0]
        except IndexError:
            logger.exception('Operation provided to convert to cli has no action. ex. "mkhost".')
            raise
        return action

    @staticmethod
    def _get_operation_params(shlexd_body):
        operation_params_list = shlexd_body[1:]
        params = ActInput._parm_list_to_dict(operation_params_list)
        return params

    @property
    def act_method(self):
        if not self._act_method:
            self.parse_operation()
        return self._act_method

    @act_method.setter
    def act_method(self, act_method):
        if act_method == "udsinfo":
            self.http_method = "get"
            self.base_path = "info"
        elif act_method == "udstask":
            self.http_method = "post"
            self.base_path = "task"
        elif act_method == "sainfo":
            self.http_method = "get"
            self.base_path = "shinfo"
        else:
            raise ValueError("Actifio CLI API doesn't support method '{}'".format(act_method))

        self._act_method = act_method

    # Parse operation properties
    # Avoid writing the below a bunch of times:
    # @property
    # def http_method(self):
    #     if not self._http_method:
    #         self.parse_operation()
    #     return self._http_method
    #
    # @http_method.setter
    # def http_method(self, value):
    #     self._http_method = value

    @staticmethod
    def _make_parse_op_prop(prop_name):
        private_var_name = '_' + prop_name

        def fget(self):
            if not getattr(self, private_var_name):
                self.parse_operation()
            return getattr(self, private_var_name)

        def fset(self, value):
            setattr(self, private_var_name, value)

        return property(fget, fset)

    for parse_op_prop in parse_op_props:
        # This relies on cpython's implementation aka PEP558 https://www.python.org/dev/peps/pep-0558/#class-scope
        locals()[parse_op_prop] = _make_parse_op_prop.__func__(parse_op_prop)


class ActOutput(Act_put):
    def __init__(self):
        self._result = {}
        self._stdout = self._stderr = self._sshout = None

    @property
    def result(self):
        """
        The result in a dictionary how a RESTful request would respond.

        We remove 'result' to fix the case where stdout was loaded before stderr and we didn't know at parse time if the
        response was empty or an error.

        Without this handling we would receive:
        `{'errorcode': 010010, 'errormessage': 'invalid option: k', 'result': []}`
        Instead of the correct response:
        `{'errormessage': 'invalid option: k', 'errorcode': 10010}`
        """
        if {'result', 'errorcode', 'errormessage'} <= self._result.keys():
            empty_result = self._result.pop('result')
            logger.debug(f"Removed empty result from ActOutput with error. {empty_result}")
        return self._result

    @result.setter
    def result(self, value):
        # TODO: Make result setter smarter; allow a switch based on conn_object to fill in result or stdout and stderr
        self._result = value

    @property
    def response(self):
        ar = ActResponse(self.result)
        return ar

    @property
    def sshout(self):
        return self._sshout

    @sshout.setter
    def sshout(self, ssh_out):
        self.stderr = ssh_out.err
        self.stdout = ssh_out.out

    @property
    def stdout(self):
        return self._stdout

    @stdout.setter
    def stdout(self, stdout):
        self._stdout = stdout
        if not self.stderr:
            self.result.update(self._parse_ssh_stdout(stdout))

    @property
    def stderr(self):
        return self._stderr

    @stderr.setter
    def stderr(self, stderr):
        self._stderr = stderr
        self.result.update(self._parse_ssh_stderr(stderr))

    @staticmethod
    def _parse_ssh_stderr(stderr):
        error_dict = {}
        if stderr:
            error_code = re.match("(?:ACTERR-)(\d+)", stderr[0]).group(1)
            error_message = re.match("(?:ACTERR-\d+ )(.+)", stderr[0]).group(1)
            error_dict['errorcode'] = int(error_code)
            error_dict['errormessage'] = error_message

        return error_dict

    def _parse_ssh_stdout(self, stdout):
        result_dict = {}
        try:
            # Handle udsinfo commands that start with ls
            if self._ssh_delim in stdout[0]:
                logger.info('Ssh operation contains delim, grabbing header.')
                headers = stdout[0].split(self._ssh_delim)
                values = [row.split(self._ssh_delim) for row in stdout[1:]]
                # Make the result list like JSON response
                result_list = []
                for row in values:
                    result_list.append(dict(zip(headers, row)))
                    result_dict['result'] = result_list
            else:
                # Handle all other udsinfo commands
                # only return a list if there is more than one item, else return a string (see exception)
                if stdout[1]:
                    result_dict = {'result': stdout}
                # TODO Do we need ssh_deliming on any results that match "udsinfo -h | egrep -v '^\s*ls|*list'"
        except IndexError:
            try:
                # return single item response as a string
                result_dict['result'] = stdout[0]
            except IndexError:
                # return "[]" if stdout was empty, matching standard rest response
                result_dict['result'] = stdout

        return result_dict

    def convert_ssh_output(self, stdout, stderr):
        logger.debug('Parsing stdout: {}'.format(stdout))
        rest_like_result = {}
        # first add error info if present
        rest_like_result.update(self._parse_ssh_stderr(stderr))
        # Now format stdout
        rest_like_result.update(self._parse_ssh_stdout(stdout))
        logger.debug("Converted cli result: {}".format(rest_like_result))
        return rest_like_result


class ActCmd:
    """Abstract single CLI API command"""
    def __init__(self, conn_object=None):
        self.input = None
        self.output = ActOutput()
        self.conn_object = conn_object
        logger.debug("Actifio command started with {a} {conn} connection".format(
            a='an' if self.conn_type == 'ssh' else 'a', conn=self.conn_type)
        )

    def cmd_input(self, operation, **update_cmds):
        self.input = ActInput(operation, conn_object=self.conn_object, update_cmds=update_cmds)

    @property
    def call_kwargs(self):
        return getattr(self.input, self.conn_type + '_kwargs')

    @property
    def conn_type(self):
        from actappliance.connections import ApplianceRest, ApplianceSsh
        if isinstance(self.conn_object, ApplianceSsh):
            return 'ssh'
        elif isinstance(self.conn_object, ApplianceRest):
            return 'rest'

    @property
    def response(self):
        act_response = self.output.response
        # logger.debug(f"Actifio response:\n{act_response}")
        act_response.input = self.input
        return act_response



from actappliance.models import ActResponse

# Categorize keywords so dynamic keyword generation decisions can be made
# Basic conversion table to change CLI commands into english
keyword_prefixes = {'make': 'mk', 'add': 'mk',
                    'change': 'ch', 'set': 'ch',
                    'remove': 'rm', 'absent': 'rm',
                    'find': 'ls', 'search': 'ls',
                    'execute': '', 'do': '',
                    'info': '', 'perform': ''}

# Udstask keywords
task_keywords = ('make', 'add',
                 'change', 'set',
                 'remove', 'absent',
                 'execute', 'do')

# Udsinfo keywords
info_keywords = ('find', 'search',
                 'info', 'perform')

# Keywords that can be entirely generated
dynamic_keywords = ('make',
                    'change',
                    'remove',
                    'find',
                    'execute',
                    'info')

# Keywords that need human interaction to function. Either with error handlers if the action is not-idempotent or
# with retries if it is idempotent
idempotent_keywords = ('add',
                       'set',
                       'absent',
                       'search',
                       'do',
                       'perform')


class APIHelper(object):
    def __init__(self, appliance_obj):
        self.a = appliance_obj

    def __getattr__(self, name):
        """
        Create simple API helpers which send the call and (by default) check for errors.

        :param name: dynamic keyword
        :return: method
        """
        method = self.keyword_gen(name)

        if method:
            return self.keyword_gen(name)
        else:
            raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__.__name__, name))

    def keyword_gen(self, name):
        """
        Creates a keyword method based on it's name.

        Will only function on low level non-idempotent keywords.
        This will generate a method that sends a command, ensures no error occurred and return the unparsed
        response.

        :param name: low level keyword name
        :return: method
        """
        # the split argument is a nonplussed face (>'_')>[giff methods plz]
        keyword = name.split('_')
        keyword_prefix = keyword[0]
        if keyword_prefix.lower() in dynamic_keywords:
            if keyword_prefix.lower() in task_keywords:
                command_action = 'udstask'
            elif keyword_prefix.lower() in info_keywords:
                command_action = 'udsinfo'
            else:
                raise RuntimeError('Unsupported dynamic method. Supported methods are {}'.format(dynamic_keywords))

            # acquire the body of the action i.e. workflow in udsinfo lsworkflow
            command_body = ''.join(keyword[1:])
            # manufacture the whole command
            command = command_action + ' ' + keyword_prefixes[keyword_prefix] + command_body

            def method(*, raise_for_error=True, **update_cmds):
                r = self.a.cmd(command, **update_cmds)
                assert issubclass(type(r), ActResponse)
                if raise_for_error:
                    r.raise_for_error()

                return r

            method.__name__ = name
            setattr(self, name, method.__get__(self))

            return method



import paramiko
import atexit
from collections import namedtuple
from actappliance.connections.connection import ApplianceConnection


ssh_out = namedtuple('ssh_output', ['out', 'err', 'status'])


class ApplianceSsh(ApplianceConnection):
    def __init__(self, *args, **kwargs):
        super(ApplianceSsh, self).__init__(*args, **kwargs)

        # SSH delimiter to use during uds commands
        self.SSH_DELIM = '|MiLeD|'

        # set in method 'connect'
        self.ssh_client = None

    def connect(self):
        """
        Connect to an appliance, disable timeouts and return the ssh object.
        :return: paramiko ssh object
        """
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            ssh.connect(self.system, **self.connect_kwargs)
        except paramiko.SSHException:
            self.logger.exception("Failed to connect to {}".format(self.system))
            raise
        _, out, _ = ssh.exec_command("unset TMOUT")
        rc = out.channel.recv_exit_status()
        if rc != 0:
            raise RuntimeError("Return code of 'unset TMOUT was {}.".format(rc))
        self.ssh_client = ssh

        # Close this connection after execution
        atexit.register(self.disconnect)

    def disconnect(self):
        try:
            self.ssh_client.close()
        except (AttributeError, NameError):
            self.logger.debug("No ssh client to close.")
        self.ssh_client = None

    def _auto_connect(self):
        """
        :return: ActCmd object with ssh_command attribute
        """
        if not self.ssh_client:
            self.connect()

    def _get_ssh_out_from_command(self, exec_args=None, exec_kwargs=None, strip_newlines=False):
        exec_args = [] if exec_args is None else exec_args
        exec_kwargs = {} if exec_kwargs is None else exec_kwargs

        stdin_chan, stdout_chan, stderr_chan = self.ssh_client.exec_command(*exec_args, **exec_kwargs)
        stdin_chan.close()

        # Wait for exit_status
        exit_status = stdout_chan.channel.recv_exit_status()

        # get output from channelfiles
        stdout = stdout_chan.readlines()
        stderr = stderr_chan.readlines()
        if strip_newlines:
            stdout = [line.rstrip('\n') for line in stdout]
            stderr = [line.rstrip('\n') for line in stderr]
        self.logger.debug('stderr: \n{}'.format(stderr))
        self.logger.debug("exit status: {}".format(exit_status))
        return ssh_out(out=stdout, err=stderr, status=exit_status)

    def call(self, act_cmd):
        """
        Use when you explicitly want to connect over ssh and send raw ssh commands.

        This houses the functionality unlike rest method because I want to avoid passing around the client to retain
        channels.

        :param act_cmd: Instance of ActCmd object
        :return: stdout, stderr, exit status
        """
        super(ApplianceSsh, self).call(act_cmd)

        ssh_o = self._get_ssh_out_from_command(exec_kwargs=act_cmd.input.ssh_kwargs, strip_newlines=True)

        act_cmd.output.sshout = ssh_o

    def raw(self, ssh_command, statuses=None):
        """
        Send an ssh command with the only processing being newline stripping.

        Threaded timeouts still apply to `raw`. If you need absolute individual control use method `call`.

        :param ssh_command: command to send
        :param statuses: list of acceptable exit statuses
        :return: namedtuple ssh_output
        """
        self._auto_connect()

        ssh_o = self._get_ssh_out_from_command(exec_args=[ssh_command], strip_newlines=True)

        if statuses and ssh_o.status not in statuses:
            self.logger.debug("Failed stdout:\n{}\nFailed stderr:\n{}".format(ssh_o.out, ssh_o.err))
            raise RuntimeError("Exit status of command {} was {}; only {} accepted.".format(
                ssh_command, ssh_o.status, ','.join(str(s) for s in statuses))
            )

        return ssh_o
